from enum import Enum
from random import randint
from typing import List, final, Optional, Callable
from dataclasses import dataclass, field

from torch import distributed as dist
from torch.cuda import device_count

from .io import IO


@final
class Device(Enum):
    r"""The device on which to run the task."""

    CPU = 'cpu'
    r"""The device to run on is a CPU."""

    GPU = 'gpu'
    r"""The device to run on is one or more GPUs."""

    SLURM = 'slurm'
    r"""The device to run on is a cluster of GPU nodes managed by SLURM."""

    MPS = 'mps'
    r"""The device to run on is Apple SoC."""

    @staticmethod
    def from_str(device: str) -> 'Device':
        r"""
        This method returns a :py:class:`Device` object given a valid device
        string.

        :param str device: The type of the device. Supported values: ``cpu``,
            ``gpu``, ``slurm``, and ``mps`` (case insensitive).

        :returns Device: :py:class:`Device` corresponds to the device type
            string.

        :raises ValueError: Raises an error if the string is invalid.
        """

        match device.lower():
            case 'cpu': return Device.CPU
            case 'gpu': return Device.GPU
            case 'slurm': return Device.SLURM
            case 'mps': return Device.MPS
            case _: raise ValueError("Invalid device string specified.")


@final
@dataclass
class Platform:
    r"""
    Platform-related configurations such as the device, environment,
    communication IP address and port, world size, `etc.`

    .. admonition:: Examples
        :class: tip

        .. code:: python

            from ddpw import Platform

            # a setup with 4 GPUs
            platform = Platform(device='gpu', n_gpus=4)

            # a setup to request SLURM for 2 nodes, each with 3 GPUs in the "example" partition
            platform = Platform(device='slurm', n_nodes=2, n_gpus=3, partition='example')
    """

    name: str = 'ddpw'
    r"""Name of the platform job. Used by SLURM. Default: ``ddpw``."""

    device: Device | str = Device.GPU
    r"""The type of device. Default: ``Device.GPU``."""

    partition: str = 'general'
    r"""Name of the SLURM partition (used only by SLURM). Default: ``general``."""

    n_nodes: int = 1
    r"""The total number of nodes (used only by SLURM).  Default: ``1``."""

    n_gpus: int = 1
    r"""The number of GPUs (per node). Default: ``1``."""

    n_cpus: int = 1
    r"""The total number of CPUs (used only by SLURM). Default: ``1``."""

    ram: int = 32
    r"""Total RAM (in GB) (used only by SLURM). Default: ``32``."""

    spawn_method: Optional[str] = 'fork'
    r"""This string corresponds to that passed to :meth:`mp.set_start_method`.
    Default: ``fork``."""

    ipc_protocol: str = 'tcp'
    r"""IPC protocol. Accepted values: ``tcp`` and ``file``. Default:
    ``tcp``."""

    master_addr: str = 'localhost'
    r"""IPC address. Default: ``localhost``."""

    master_port: Optional[str] = str(randint(1024, 49151))
    r"""The port at which IPC happens. Default: a random port between 1024 and
    49151."""

    ipc_groups: Optional[List[List[int]]] = field(default_factory=lambda: [])

    r"""A list of lists of non-overlapping global ranks of devices. If ``None``,
    every device will be its own group, and no IPC will take place. If an empty
    list is passed, all devices are grouped into one process group. Default:
    ``[]``.

    .. admonition:: Examples
        :class: tip

        .. code:: python

            # no IPC between devices; each device is its own group
            platform = Platform(device='gpu', n_gpus=4, ipc_groups=None)

            # all devices under one group: default behaviour
            platform = Platform(device='gpu', n_gpus=4)
            platform = Platform(device='gpu', n_gpus=4, ipc_groups=[])

            # custom groups
            platform = Platform(device='gpu', n_gpus=4, ipc_groups=[[0, 2], [1], [3]])
            platform = Platform(device='gpu', n_gpus=4, ipc_groups=[[0, 2], [1, 3]])

    .. admonition:: Variable groups unstable
        :class: warning

        PyTorch behaviour seems to be inconsistent when using variable process
        groups. An `open bug <https://github.com/pytorch/pytorch/issues/29115>`_
        issue is on GitHub.
    """

    backend: Optional[dist.Backend] = dist.Backend.GLOO if \
        hasattr(dist, 'Backend') else None
    r"""The PyTorch-supported backend to use for distributed data parallel.
    Default: ``torch.distributed.Backend.GLOO``."""

    seed: int = 1889
    r"""Seed with which to initialise the various [pseudo]random number
    generators. Default: ``1889``."""

    timeout_min: int = 2880
    r"""Minimum timeout (in minutes) for jobs (used only by SLURM). Default:
    ``2880`` (two days)."""

    slurm_additional_parameters: Optional[dict] = None
    r"""Additional SLURM parameters; this dictionary corresponds to the one
    passed to ``submitit``'s ``slurm_additional_parameters`` `argument
    <https://github.com/facebookincubator/submitit/issues/23>`_. Default:
    ``None``."""

    console_logs: str = './logs'
    r"""Location of console logs (used mainly by SLURM to log the errors and
    output to files). Default: ``./logs``"""

    verbose: Optional[bool] = True
    r"""Whether or not to print updates to the standard output during setup.
    Default: ``True``."""

    upon_finish: Optional[Callable] = None
    r"""An optional callable to be invoked upon completion of the given task.
    Default: ``None``."""

    def __post_init__(self):
        if isinstance(self.device, str):
            self.device = Device.from_str(self.device)

    @property
    def world_size(self):
        r"""Specified the world size. This is the total number of GPUs across
        all nodes. Default: ``1``."""

        n_nodes = self.n_nodes if self.device == Device.SLURM else 1 
        n_gpus = min(self.n_gpus, device_count()) \
                if self.device == Device.GPU else self.n_gpus

        return n_nodes * n_gpus

    @property
    def requires_ipc(self):
        r"""Specified whether the processes need inter-communication. This
        property determines whether or not the setup requires IPC. IPC is not
        required for a single device."""

        if self.device in [Device.CPU, Device.MPS]: return False

        if self.ipc_groups is None: return False

        if self.device == Device.GPU:
            return device_count() > 1 and self.world_size > 1

        return self.world_size > 1

    def print(self):
        r"""
        This method serialises this object in a human readable format and
        prints it.
        """

        details = f"""
        \r Platform details:

        \r • Name:\t\t\t\t{self.name}
        \r • Device:\t\t\t\t{self.device.value.upper()}
        \r • CPUs:\t\t\t\t{self.n_cpus}
        \r • Total RAM:\t\t\t\t{self.ram}GB
        \r • GPUs (per node):\t\t\t{self.n_gpus} (requested)
        \r • GPUs (per node):\t\t\t{device_count()} (available)
        \r • PyTorch backend:\t\t\t{self.backend}
        \r • Seed (random number generators):\t{self.seed}
        """

        if self.device == Device.SLURM or True:
            details += f"""\
            \r • Nodes:\t\t\t\t{self.n_nodes}
            \r • SLURM partition:\t\t\t{self.partition}
            \r • SLURM timeout:\t\t\t{self.timeout_min} minutes
            """

        if not self.requires_ipc:
            details += f"""\
            \r • Spawn method:\t\t\t{self.spawn_method}
            \r • Master IP address:\t\t\t{self.master_addr}
            \r • Master port:\t\t\t\t{self.master_port}
            """

        details += f"""\
        \r • World size:\t\t\t\t{self.world_size}
        """

        IO.print(details)




from typing import Optional


class IO:
    verbose: Optional[bool] = True
    r"""A global boolean property that specifies if the wrapper must output
    updates to the console or not."""

    @staticmethod
    def print(*args, **kwargs):
        r"""
        A custom print wrapper that outputs the contents if the process is
        running in the verbose mode (`i.e.`, ``verbose = True``). This method is
        a simple check around Python's system print function.

        :param bool verbose: To print or not to print. Default: ``None``.
        """

        if 'flush' not in kwargs:
            kwargs['flush'] = True

        if kwargs.get('verbose', IO.verbose):
            if 'verbose' in kwargs: del(kwargs['verbose'])
            print(*args, **kwargs)




from random import seed as rand_seed
from typing import Optional

from numpy.random import seed as np_seed
from torch import Tensor, device as t_device, distributed as dist
from torch.random import manual_seed
from torch.cuda import manual_seed_all, is_available, current_device
from torch.nn import (Module, BatchNorm1d, BatchNorm2d, BatchNorm3d,
                      SyncBatchNorm)
from torch.optim import Optimizer
from torch.nn.parallel import DistributedDataParallel
from torch.utils.data import Dataset, DistributedSampler
 
from .platform import Device, Platform


def seed_generators(seed: int):
    r"""
    This function seeds [pseudo]random number generators from various packages.

    :param int seed: The seed.
    """

    rand_seed(seed)
    np_seed(seed)
    manual_seed(seed)
    manual_seed_all(seed)


def average_params_grads(module: Module, params: bool = False,
                         grads: bool = True):
    r"""
    Given a module, this function averages the parameters of the given model
    and/or their gradients across all the GPUs (copied over from a `PyTorch blog
    <https://pytorch.org/tutorials/intermediate/dist_tuto.html>`_ and further
    modified here).

    :param torch.nn.Module module: The module whose parameters/gradients are to
        be averaged.
    :param bool params: Whether to average the parameters or not. Default:
        ``False``.
    :param bool grads: Whether to average the gradients or not. Default:
        ``True``.
    """

    if not (params and grads): return

    world_size = float(dist.get_world_size())

    for p in module.parameters():
        if p.grad is not None and grads:
            dist.all_reduce(p.grad, op=dist.ReduceOp.SUM)
            p.grad /= world_size
        if p is not None and params:
            dist.all_reduce(p, op=dist.ReduceOp.SUM)
            p /= world_size


def optimiser_to(optimiser: Optimizer, device: t_device):
    r"""
    This function offers a simple way to move all the parameters optimised by an
    optimiser to the specified device. This function has been taken as is from a
    `solution on PyTorch Discuss
    <https://discuss.pytorch.org/t/moving-optimizer-from-cpu-to-gpu/96068/3>`_.

    :param torch.optim.Optimizer optimiser: The optimiser to move to a device.
    :param torch.device device: The device to which to move the optimiser.
    """

    for param in optimiser.state.values():
        # if any global tensors in the state dict
        if isinstance(param, Tensor):
            param.data = param.data.to(device)
            if param._grad is not None:
                param._grad.data = param._grad.data.to(device)
        elif isinstance(param, dict):
            for subparam in param.values():
                if isinstance(subparam, Tensor):
                    subparam.data = subparam.data.to(device)
                    if subparam._grad is not None:
                        subparam._grad.data = subparam._grad.data.to(device)


def has_batch_norm(module: Module) -> bool:
    r"""
    This function checks if a module has batch normalisation layer(s) in it.

    :param nn.Module module: The module to be checked for containing any batch
        normalisation layers.

    :returns bool: Whether or not the module has batch normalisation layer(s) in
        it.
    """

    if isinstance(module, (BatchNorm1d, BatchNorm2d, BatchNorm3d)):
        return True

    for child in module.children():
        if has_batch_norm(child): return True

    return False


def to(module: Module, local_rank: int, sync_modules: bool = True,
       device: Device = Device.GPU) -> Module:
    r"""
    A quick and minimal function that works on all devices to move the given
    module to the specified device: CPU, GPU, or MPS. If the platform is set up
    in an IPC fashion, this function optionally moves the module in a
    distributed data parallel-fashion and synchronises batch normalisation
    layers, if any.

    :param torch.nn.Module module: The module to be moved.
    :param int local_rank: The local rank of the device.
    :param bool sync_modules: Whether to synchronise the modules across devices
        or not. If yes, the module becomes ``DistributedDataParallel``. Default:
        ``True``.
    :param Device device: The type of device to which to move the module. This
        argument is useful because once ``device`` has been globally set, this
        function can be called regardless of the current device in which the
        items are and thus helps avoid additional checks. Default:
        ``Device.GPU``.

    :returns torch.nn.Module: The module moved to the appropriate device.
    """

    match device:
        case Device.CPU | Device.MPS:
            module = module.to(device.value)
        case _:
            module = module.cuda(t_device(local_rank))

    if sync_modules and device not in [Device.CPU, Device.MPS]:
        if has_batch_norm(module):
            module = SyncBatchNorm.convert_sync_batchnorm(module)

        if dist.is_initialized():
            module = DistributedDataParallel(module, device_ids=[local_rank],
                                             find_unused_parameters=False)

    return module


def get_dataset_sampler(dataset: Dataset, global_rank: int,
                        platform: Platform) -> Optional[DistributedSampler]:
    r"""
    This function selects a portion of the original dataset shared by other
    devices. If the device is CPU or MPS, no sharing is necessary.

    :param data.Dataset dataset: The dataset from which to sample for the
        current device.
    :param int global_rank: The global rank of the device.
    :param Platform platform: Platform-related configurations.

    :returns DistributedSampler: Dataset sampler for the given dataset and
        world size.
    """

    sampler = None

    if platform.device not in [Device.CPU, Device.MPS]:
        sampler = DistributedSampler(dataset, num_replicas=platform.world_size,
                                     rank=global_rank)

    return sampler

 
def device(module: Module) -> t_device:
    r"""
    Given a module, this function returns the device on which it currently
    resides. If the module has no parameters, the current device is returned by
    default.

    :param nn.Module module: The module whose device is sought.

    :returns torch.device: The device of the module.
    """

    from torch.backends.mps import is_available as is_mps_available

    p = module.parameters()
    try:
        device = next(p).device
    except StopIteration:
        if is_available(): device = t_device(f'cuda:{current_device()}')
        elif is_mps_available(): device = t_device('mps')
        else: device = t_device('cpu')

    return device




from datetime import timedelta
from os import environ
from os.path import isabs, abspath, expanduser
from typing import Any, Callable, Optional, Tuple

from torch.distributed import (init_process_group, destroy_process_group,
                               new_group, barrier, GroupMember, ProcessGroup)
from torch.multiprocessing import set_start_method, Process

from .platform import Device, Platform
from .io import IO
from . import functional as DF


def setup(node: int, global_rank: int, local_rank: int, platform: Platform,
          target: Callable[[int, int, ProcessGroup, Optional[Tuple]], Any],
          args: Optional[Tuple]) :
    r"""
    This function is called at the beginning of the process in each device
    (CPU/GPU). Depending on the needs, this function establishes DDP
    communication protocols, seeds random number generators, invokes the given
    task, and performs cleanup tasks.

    :param int node: Node number.
    :param int global_rank: Global rank of the device.
    :param int local_rank: Local rank of the device.
    :param Platform platform: Platform-related configurations.
    :param Callable target: The callable task to invoke upon finishing setup.
    :param Optional[Tuple] args: Arguments to be passed to ``target``.
    """

    node_info = f'Node {node}, GPU {global_rank}(G)/{local_rank}(L)'

    IO.print(f'[{node_info}] Initialising the process.')

    if platform.requires_ipc:
        environ['MASTER_ADDR'] = platform.master_addr
        port = environ['MASTER_PORT'] = str(platform.master_port)

        im = f'{platform.ipc_protocol}://{environ["MASTER_ADDR"]}'
        if port is not None: im = f'{im}:{environ["MASTER_PORT"]}'

        IO.print(f'[{node_info}] IPC at {im}.')

        # do not specify the rank if process groups are to be used
        init_process_group(backend=platform.backend, init_method=im,
                           rank=global_rank, world_size=platform.world_size)

    # 1. Seed random number generators
    IO.print(f'[{node_info}] ' +
                f'Seeding random number generators with {platform.seed}.')
    DF.seed_generators(platform.seed)

    # organise groups
    grp = GroupMember.WORLD
    if platform.requires_ipc and len(platform.ipc_groups) > 0:
        device_group = grp
        for device_group in platform.ipc_groups:
            if global_rank in device_group: break
        grp = new_group(ranks=device_group, timeout=timedelta(seconds=30),
                        backend=platform.backend)

    # 2. Wait for all the processes in this group to synchronise and then start
    # the task
    if platform.requires_ipc: barrier(grp)

    # 3. Invoke the given task
    IO.print(f'[{node_info}] All setup finished.')
    target(global_rank, local_rank, grp, args)

    # 4. Cleanup
    if platform.requires_ipc: destroy_process_group(grp)
    IO.print(f'[{node_info}] Tasks on device complete.')
 

class Wrapper:
    r"""
    This class is the highest level of abstraction: it accepts the
    platform-related configurations and initialises the setup accordingly. When
    given a task, it then runs the task according to the specified
    configurations.

    .. admonition:: Example
        :class: tip

        .. code:: python

            from ddpw import Platform, Wrapper

            wrapper = Wrapper(Platform(...))

            wrapper.start(some_callable)

    :param Platform platform: Platform-related configurations.
    """

    def __init__(self, platform: Platform):
        IO.verbose = platform.verbose

        IO.print('Initialising the DDP Wrapper.')
        self.platform = platform

        if platform.requires_ipc:
            try:
                set_start_method(platform.spawn_method)
            except RuntimeError as e:
                IO.print(
                  f'Warning: {e}. Skipping setting the start method for forks.')

    def __gpu(self, target: Callable[[int, int, ProcessGroup,
                              Optional[Tuple]], Any], args: Optional[Tuple]):
        r"""
        This method spins up a process for each GPU in the world. It assigns the
        task to be run on each process, `viz.`, distributing the datasets and
        models and commencing the task.

        :param Callable target: The function to call on each GPU upon setup.
        :param Optional[Tuple] args: Arguments to be passed to ``target``.
        """

        if self.platform.world_size == 1:
            node_info = f'Node 0, GPU 0(G)/0(L)'
            IO.print(f'[{node_info}] Task starting on GPU.')
            setup(0, 0, 0, self.platform, target, args)
            return

        IO.print(f'Spawning {self.platform.world_size} processes.')
        processes = []

        # create a process for each GPU in the world
        for rank in range(self.platform.world_size):
            p = Process(target=setup, args=(0, rank, rank, self.platform,
                                               target, args))
            processes.append(p)
            p.start()

        for p in processes: p.join()

        IO.print('All processes complete.')

    def __slurm(self, target: Callable, console_logs: str):
        r"""
        Similar to :py:meth:`.__gpu` but for SLURM. An additional step includes
        spinning up a process for each node, done with ``submitit``.

        :param Callable target: The function to call on each GPU upon setup.
        :param str console_logs: Location to save SLURM console logs.
        """
        from submitit import AutoExecutor

        IO.print('Setting up the SLURM platform.')

        executor = AutoExecutor(folder=console_logs)
        executor.update_parameters(
            name=self.platform.name,
            mem_gb=self.platform.ram,
            gpus_per_node=self.platform.n_gpus,
            tasks_per_node=self.platform.n_gpus,
            cpus_per_task=self.platform.n_cpus,
            nodes=self.platform.n_nodes,
            timeout_min=self.platform.timeout_min,
            slurm_partition=self.platform.partition
        )

        if self.platform.slurm_additional_parameters is not None:
            executor.update_parameters(slurm_additional_parameters=
                                   self.platform.slurm_additional_parameters)

        return executor.submit(target)

    def start(self, target: Callable[[int, int, ProcessGroup, Optional[Tuple]],
                                     Any], args: Optional[Tuple] = None):
        r"""
        This method performs the necessary setup according to the specified
        configurations and then invokes the given task.

        :param Callable[[int, int, dist.ProcessGroup, Optional[Tuple]], Any] target:
            The task. A
            callable which accepts two integers (the global
            and local ranks of the device), the process group, and an optional
            tuple which are the callable's arguments.
        :param Optional[Tuple] args: Arguments to be passed to ``target``.
            Default: ``None``.
        """

        self.platform.print()

        IO.print('Starting process(es).')

        def finished():
            if self.platform.upon_finish is not None and \
                callable(self.platform.upon_finish):
                return self.platform.upon_finish()

        match self.platform.device:
            case Device.CPU | Device.MPS:
                setup(0, 0, 0, self.platform, target, args)
                finished()
            case Device.GPU:
                self.__gpu(target, args)
                finished()
            case Device.SLURM:
                def individual_gpu():
                    r"""
                    This nested function is the starting point for each
                    SLURM-based GPU.
                    """
                    from submitit import JobEnvironment

                    self.platform.master_addr = environ['HOSTNAME']
                    job_env = JobEnvironment()

                    details = f"""
                    \r • Node: {job_env.node}.
                    \r • Global rank: {job_env.global_rank}.
                    \r • Local rank: {job_env.local_rank}.
                    """
                    IO.print(details)

                    setup(job_env.node, job_env.global_rank, job_env.local_rank,
                          self.platform, target, args)

                    if job_env.global_rank == 0: finished()

                job = self.__slurm(individual_gpu, self.platform.console_logs)

                p = self.platform.console_logs
                if not isabs(p): p = abspath(expanduser(p))
                details = f"""
                \rSLURM job "{self.platform.name}" ({job.job_id}) scheduled.
                \rSee respective device logs for output on those devices.
                \rLogs saved at {p}.
                """

                print(details)


def wrapper(platform: Platform):
    r"""
    A decorator that can be applied to callables.

    :param Platform platform: Platform details.

    .. admonition:: Example
        :class: tip

        .. code:: python

            from ddpw import Platform, wrapper

            @wrapper(Platform(device='gpu', n_gpus=2, n_cpus=2))
            def run(a, b):
                # some task
                pass
    """

    def __ddpw(fn):
        def __wrapper(*args, **kwargs):
            def __my_fn(global_rank, local_rank, _, __):
                return fn(*args, **kwargs)

            Wrapper(platform).start(__my_fn)

        return __wrapper

    return __ddpw

